--euparse.e
--Given a Source Object (created with poeparse.e)
--extract definitions and declarations.
--Note: THE SOURCE FILE MUST BE CORRECT!
--	If the source file has errors, you may get strange results.
--	But mistakes that would cause corruption would be like
--      * defining a global inside a code block
--	* defining a type inside a function, type, procedure..
--	* redefining a function, type or procedure globally.

--	Accept for the last one, these are all pretty rare.


--Requires srcobj.e

--*************************************************
--If you decide to add functionality, such as counting how often a function
--is called, email me and I'll add it to the official version.

--Information will be stored in the following semi-sets:
--File Set
--Global Type Set
--Global Function Set
--Global Procedure Set
--Global Variable Sets
--All Types Set
--Toplevel Variable Sets
--Type Variable Sets		(defined inside a type)
--Function Variable Sets	(defined inside a Function)
--Procedure Variable Sets	(defined inside a Procedure)

--The data will be stored as follows:
--(square brackets means indexed by)
--File Set
--{"File Name", ... }

--Global Type Set
--{"Type Name", ... }

--Global Function Set
--{"Function Name", ... }

--Global Procedure Set
--{"Procedure Name", ... }

--Global Constant Set
--{"Constant Name", ... }

--Global Variable Set [All Types Set]
--{{"Variable Name", ... }, ... }

--All Types Set
--{"Type Name", ... }

--Type Sets [File Set]
--{{"Type Name", ... }, ... }

--Function Sets [File Set]
--{{"Function Name", ... }, ... }

--Procedure Sets [File Set]
--{{"Procedure Name", ... }, ... }

--Constant Sets [File Set]
--{{"Constant Name", ... }, ... }

--Toplevel Variable Sets [File Set] [All Types Set]
--{{{"Variable Name", ... }, ... }, ... }

--Type Variable Sets [File Set] [Type Sets] [All Types Set]
--{{{"Variable Name", ... }, ... }, ... }

--Function Variable Sets [File Set] [Function Sets] [All Types Set]
--{{{"Variable Name", ... }, ... }, ... }

--Procedure Variable Sets [File Set] [Procedure Sets] [All Types Set]
--{{{"Variable Name", ... }, ... }, ... }

sequence SLASH
--Platform stuff
if platform() > 2 then
  SLASH = "/"
else
  SLASH = "\\"
end if

--Set include dir
constant EU_INCLUDE_DIR = getenv("EUDIR") & SLASH & "include"
--include files
include semisetobj.e
include srcobj.e

--Define Tables
global SemiSet File_Set,
               Global_Type_Set,
	       Global_Function_Set,
	       Global_Procedure_Set,
       	 Global_Variable_Set,
             Global_Constant_Set,
	       All_Types_Set,
  	       Type_Sets,
             Constant_Sets,
	       Function_Sets,
	       Procedure_Sets,
	       Toplevel_Variable_Sets,
	       Type_Variable_Sets,
	       Function_Variable_Sets,
	       Procedure_Variable_Sets

--Initialize the tables
File_Set = {}
Global_Type_Set = {"atom","integer","sequence","object"}
Global_Constant_Set = {}
Global_Function_Set = {}
Global_Procedure_Set = {}
Global_Variable_Set = {{},{},{},{}}
All_Types_Set = {"atom","integer","sequence","object"}
Type_Sets = {}
Constant_Sets = {}
Function_Sets = {}
Procedure_Sets = {}
Toplevel_Variable_Sets = {}
Type_Variable_Sets = {}
Function_Variable_Sets = {}
Procedure_Variable_Sets = {}

function flatten_source(sequence src)
  sequence new_src
  new_src = {}
  if length(src) then
    for i=1 to length(src) do
      if length(src[i]) then
        for j=1 to length(src[i]) do
	  new_src &= {src[i][j]}
	end for
      end if  
    end for
  end if 
  return new_src
end function  

function handle_constants(atom w_ind, sequence so, integer global_flag)
  integer parens,
          atomic_count,
	  file_ind
  sequence c_word
  
  file_ind = length(File_Set)
  while w_ind <= length(so) do
    c_word = so[w_ind]
    if global_flag then
      Global_Constant_Set = SemiSet_Insert(Global_Constant_Set, c_word)
    end if
    Constant_Sets[file_ind] = SemiSet_Insert(Constant_Sets[file_ind], c_word)
    w_ind += 2 -- look at expression
    if w_ind <= length(so) then
      --skip the expression
      parens = 0
      atomic_count = 0
      --The algorithm:
      --  If two atomic expressions occur, the expression is over 
      --  or a , occurs and the paren count is zero then another constant
      while w_ind <= length(so) do
        c_word = so[w_ind]
        if find(c_word, {"{","[","("}) then
	  parens += 1
	elsif find(c_word, {"}","]",")"}) then
	  parens -= 1
	elsif find(c_word, {"=","+","-","*","/",">",">=","<","<=","and",
	                    "or","not","&","!="}) then
	  atomic_count = 0
	elsif equal(c_word,",") then
	  if parens < 1 then
	    exit
	  end if
	  atomic_count = 0 --{ foo , bar }
	else --current word is an atomic expression
	  atomic_count += 1
	  if atomic_count > 1 then
	    return w_ind - 1
	  end if        
	end if             
	w_ind += 1  
      end while
    else
      exit
    end if
    w_ind += 1
  end while
  return w_ind
end function  



with trace


--Operates on a Source Object (given a file name)
-- and thus has a File_ prefix
-- the current algorithm will ignore words that aren't related to a
-- definition.  This means the information will still be extracted if
-- the source code contains an error (So long as 'so' is a proper Source
-- Object.)
-- Use New_Source_Object( ) to create a proper Source Object
global procedure File_Aquire_Info(sequence file_name)
  integer global_flag,
          w_ind,
          in_proc_flag,
	  in_type_flag,
	  in_func_flag,
	  in_args_flag,
	  file_ind,
	  set_ind,
	  type_ind,
          temp
  sequence c_word, 
           so,
	   temp_fname


  --First check if full path.. then extract filename.
  if find(SLASH[1],file_name) then --strip path
    temp_fname = file_name[length(EU_INCLUDE_DIR) + 2 .. length(file_name)]
  else
    temp_fname = file_name
  end if
  
  --If file_name is in File Set, done
  --no path info is stored...
  if find(temp_fname, File_Set) then
    return
  end if
  

  --Initialize Flags
  global_flag = 0
  in_proc_flag = 0
  in_type_flag = 0
  in_func_flag = 0
  in_args_flag = 0

  --Initialize Source Object
  so = New_Source_Object(file_name)
  so = flatten_source(so[SOURCE])

  file_name = temp_fname --just in case I want to use this later...

  --Insert file_name in File Set

  File_Set = SemiSet_Insert(File_Set, file_name)
  file_ind = length(File_Set) --current fileName index
  
  --Create new sets
  Type_Sets = SemiSet_Append(Type_Sets, New_SemiSet_Object())
  Function_Sets = SemiSet_Append(Function_Sets, New_SemiSet_Object())
  Procedure_Sets = SemiSet_Append(Procedure_Sets, New_SemiSet_Object())
  Constant_Sets = SemiSet_Append(Constant_Sets, New_SemiSet_Object())

  Toplevel_Variable_Sets = SemiSet_Append(Toplevel_Variable_Sets,
  			   repeat(New_SemiSet_Object(), length(All_Types_Set)))

  Type_Variable_Sets = SemiSet_Append(Type_Variable_Sets,New_SemiSet_Object())

  Function_Variable_Sets = SemiSet_Append(Function_Variable_Sets,New_SemiSet_Object())
  Procedure_Variable_Sets = SemiSet_Append(Procedure_Variable_Sets,
                                           New_SemiSet_Object())
					   
  --for each word
  if length(so) then
    w_ind = 1
    while w_ind <= length(so) do
      c_word = so[w_ind]
      if equal(c_word, "include") then
        w_ind += 1
	c_word = so[w_ind]
	temp = open(c_word,"r")
	if temp != -1 then
	  close(temp)
	  File_Aquire_Info(c_word)
	else --check include directory
	  temp = open(EU_INCLUDE_DIR & SLASH & c_word, "r")
          if temp != -1 then
            close(temp)
            File_Aquire_Info(EU_INCLUDE_DIR & SLASH & c_word)
          end if  
	end if
      elsif equal(c_word, "type") then
        w_ind += 1
	c_word = so[w_ind]	  
	if find(c_word, All_Types_Set) < 1 then --type has not been defined
          --There is now another type that may define variables at the toplevel
	  --for each existing file
	  for a=1 to length(File_Set) do
	    Toplevel_Variable_Sets[a] = SemiSet_Append(
	       Toplevel_Variable_Sets[a], New_SemiSet_Object())
	  end for
	  All_Types_Set = SemiSet_Append(All_Types_Set, c_word)  
	  if global_flag then
	    Global_Type_Set = SemiSet_Insert(Global_Type_Set, c_word)
	    global_flag = 0
	  end if 


	  --There is a new type of global variable that may be defined
	  Global_Variable_Set = SemiSet_Append(Global_Variable_Set,
	                        New_SemiSet_Object())
        end if
        Type_Sets[file_ind] = SemiSet_Append(Type_Sets[file_ind], c_word)	
        --Add an entry for this type code block   	    
	Type_Variable_Sets[file_ind] = SemiSet_Append(
	    Type_Variable_Sets[file_ind],
	    repeat(New_SemiSet_Object(),length(All_Types_Set)))
	set_ind = length(Type_Variable_Sets[file_ind]) --for variables defined inside
	in_type_flag = 1 --inside a type definition
	in_args_flag = 1
	trace(0)
      elsif equal(c_word, "function") then
        w_ind += 1
	c_word = so[w_ind]
	if global_flag then
	  Global_Function_Set = SemiSet_Insert(Global_Function_Set, c_word)
	  global_flag = 0
	end if  
	Function_Sets[file_ind] = SemiSet_Insert(Function_Sets[file_ind],
	                          c_word)
	set_ind = length(Function_Sets[file_ind]) --for vars inside
	in_func_flag = 1
	in_args_flag = 1
	--Add an entry for this function code block   	    
	Function_Variable_Sets[file_ind] = SemiSet_Append(
	    Function_Variable_Sets[file_ind],
	    repeat(New_SemiSet_Object(),length(All_Types_Set)))
      elsif equal(c_word, "procedure") then
        w_ind += 1
	c_word = so[w_ind]
	if global_flag then
	  Global_Procedure_Set = SemiSet_Insert(Global_Procedure_Set, c_word)
	  global_flag = 0
	end if  
	Procedure_Sets[file_ind] = SemiSet_Insert(Procedure_Sets[file_ind],
	                           c_word)
	set_ind = length(Procedure_Sets[file_ind])
	in_proc_flag = 1
	in_args_flag = 1
	--Add an entry for this Procedure code block   	    
	Procedure_Variable_Sets[file_ind] = SemiSet_Append(
	    Procedure_Variable_Sets[file_ind],
	    repeat(New_SemiSet_Object(),length(All_Types_Set)))
      elsif equal(c_word, "global") then
        global_flag = 1
      elsif equal(c_word, "constant") then
	w_ind = handle_constants(w_ind + 1, so, global_flag)
      elsif equal(c_word, "end") then
        w_ind += 1
	c_word = so[w_ind]
	if equal(c_word, "type") then
	  in_type_flag = 0
	elsif equal(c_word, "function") then
	  in_func_flag = 0
	elsif equal(c_word, "procedure") then
	  in_proc_flag = 0
	end if
      elsif equal(c_word, ")") then
        in_args_flag = 0	
      elsif find(c_word, All_Types_Set) then --var is being declared
        type_ind = find(c_word, All_Types_Set)
	w_ind += 1
        while w_ind <= length(so) do
	  c_word = so[w_ind]
	  if not equal(c_word, "(") then --sequence (foo)
	    if global_flag then
	      Global_Variable_Set[type_ind] =
	          SemiSet_Insert(Global_Variable_Set[type_ind], c_word)
	    elsif in_type_flag then --inside a type def
	      Type_Variable_Sets[file_ind][set_ind][type_ind] =
	      SemiSet_Insert(Type_Variable_Sets[file_ind][set_ind][type_ind],
	      c_word)
	    elsif in_func_flag then --inside a function def
	      Function_Variable_Sets[file_ind][set_ind][type_ind] =
	      SemiSet_Insert(Function_Variable_Sets[file_ind][set_ind][type_ind],
	      c_word)
	    elsif in_proc_flag then --inside a procedure def
	      Procedure_Variable_Sets[file_ind][set_ind][type_ind] =
	      SemiSet_Insert(
	          Procedure_Variable_Sets[file_ind][set_ind][type_ind],c_word)
	    else --defined at the toplevel
	      Toplevel_Variable_Sets[file_ind][type_ind] =
	      SemiSet_Insert(Toplevel_Variable_Sets[file_ind][type_ind], c_word)
	    end if
	  end if
	  if w_ind + 1 <= length(so) then
	    c_word = so[w_ind + 1] --look ahead
	    if not equal(c_word,",") then --done
	      global_flag = 0
	      exit
	    elsif not in_args_flag then --keep going
	      w_ind += 1
	    else --in arguments (done)
	      exit  
	    end if
	  end if 
	  w_ind += 1
	end while
      end if
      w_ind += 1
    end while
  end if    
end procedure
