--eusetqry.e
--Query sets defined be File_Aquire_Info() in euparse.e

--REQUIRES 
--euparse.e
--Given a Source Object (created with poeparse.e)
--extract definitions and declarations.
--Note: THE SOURCE FILE MUST BE CORRECT!
--	If the source file has errors, you may get strange results.
--	But mistakes that would cause corruption would be like
--      * defining a global inside a code block
--	* defining a type inside a function, type, procedure..
--	* redefining a function, type or procedure globally.

--	Accept for the last one, these are all pretty rare.


--*************************************************
--If you decide to add functionality, such as counting how often a function
--is called, email me and I'll add it to the official version.

--Information will be stored in the following semi-sets:
--File Set
--Global Type Set
--Global Function Set
--Global Procedure Set
--Global Variable Sets
--All Types Set
--Toplevel Variable Sets
--Type Variable Sets		(defined inside a type)
--Function Variable Sets	(defined inside a Function)
--Procedure Variable Sets	(defined inside a Procedure)

--The data will be stored as follows:
--(square brackets means indexed by)
--File Set
--{"File Name", ... }

--Global Type Set
--{"Type Name", ... }

--Global Function Set
--{"Function Name", ... }

--Global Procedure Set
--{"Procedure Name", ... }

--Global Variable Set [All Types Set]
--{{"Variable Name", ... }, ... }

--All Types Set
--{"Type Name", ... }

--Type Sets [File Set]
--{{"Type Name", ... }, ... }

--Function Sets [File Set]
--{{"Function Name", ... }, ... }

--Procedure Sets [File Set]
--{{"Procedure Name", ... }, ... }

--Toplevel Variable Sets [File Set] [All Types Set]
--{{"Variable Name", ... }, ... }, ... }

--Type Variable Sets [File Set] [Type Sets] [All Types Set]
--{{{{"Variable Name", ... }, ... }, ... }, ... }

--Function Variable Sets [File Set] [Function Sets] [All Types Set]
--{{{{"Variable Name", ... }, ... }, ... }, ... }

--Procedure Variable Sets [File Set] [Procedure Sets] [All Types Set]
--{{{{"Variable Name", ... }, ... }, ... }, ... }


--include files
include euparse.e

--Boolean Queries
global function File_Is_Included(sequence file_name)
  return find(file_name, File_Set) > 0
end function

global function Type_Is_Defined(sequence type_name)
  return find(type_name, All_Types_Set) > 0
end function

global function Function_Is_Defined(sequence func_name)
  for i=1 to length(File_Set) do
    if find(func_name, Function_Sets[i]) then
      return 1
    end if
  end for    
  return find(func_name, Global_Function_Set) > 0
end function

global function Procedure_Is_Defined(sequence proc_name)
  for i=1 to length(File_Set) do
    if find(proc_name, Procedure_Sets[i]) then
      return 1
    end if
  end for    
  return find(proc_name, Global_Procedure_Set) > 0
end function

global function Variable_Is_Defined(sequence var_name)
  for i=1 to length(File_Set) do
    for j=1 to length(All_Types_Set) do
      if find(var_name, Toplevel_Variable_Sets[i][j]) then
        return 1
      end if
      if length(Type_Sets[i]) then
        for k=1 to length(Type_Sets[i]) do
	  if find(var_name, Type_Variable_Sets[i][j][k]) then
	    return 1
	  end if
	end for
      end if
      if length(Function_Sets[i]) then
        for k=1 to length(Function_Sets[i]) do
	  if find(var_name, Function_Variable_Sets[i][j][k]) then
	    return 1
	  end if
	end for
      end if
      if length(Procedure_Sets[i]) then
        for k=1 to length(Procedure_Sets[i]) do
	  if find(var_name, Procedure_Variable_Sets[i][j][k]) then
	    return 1
	  end if
	end for
      end if
      if find(var_name,Global_Variable_Set[j]) then
        return 1
      end if
    end for              	    
  end for
  return 0
end function


--Specific Type Queries 
--(Boolean)
global function Global_Type_Is_Defined(sequence type_name)
  return find(type_name, Global_Type_Set) > 0
end function  

global function Is_Type_Defined_In(sequence type_name, sequence file_name)
  integer file_ind
  file_ind = find(file_name, File_Set)
  if file_ind then
    return find(type_name, Type_Sets[file_ind]) > 0
  end if
  return 0
end function

--(string)
global function Type_Is_Defined_In(sequence type_name)
  for i=1 to length(File_Set) do
    if find(type_name, Type_Sets[i]) then
      return File_Set[i]
    end if
  end for
  return ""    
end function


--Specific Function Queries
--(Boolean)

--(Boolean)
global function Global_Function_Is_Defined(sequence func_name)
  return find(func_name, Global_Function_Set) > 0
end function  

global function Is_Function_Defined_In(sequence func_name, sequence file_name)
  integer file_ind
  file_ind = find(file_name, File_Set)
  if file_ind then
    return find(func_name, Function_Sets[file_ind]) > 0
  end if
  return 0
end function

--(string)
global function Function_Is_Defined_In(sequence func_name)
  for i=1 to length(File_Set) do
    if find(func_name, Function_Sets[i]) then
      return File_Set[i]
    end if
  end for
  return ""    
end function

--Specific Procedure Queries
--(Boolean)
global function Global_Procedure_Is_Defined(sequence proc_name)
  return find(proc_name, Global_Procedure_Set) > 0
end function  

global function Is_Procedure_Defined_In(sequence proc_name, sequence file_name)
  integer file_ind
  file_ind = find(file_name, File_Set)
  if file_ind then
    return find(proc_name, Procedure_Sets[file_ind]) > 0
  end if
  return 0
end function

--(string)
global function Procedure_Is_Defined_In(sequence proc_name)
  for i=1 to length(File_Set) do
    if find(proc_name, Procedure_Sets[i]) then
      return File_Set[i]
    end if
  end for
  return ""    
end function

--Specific Variable Queries
--(Boolean)

global function Get_Variables_Of_Type(sequence type_name)
  integer type_ind
  sequence vars
  vars = {}
  type_ind = find(type_name,All_Types_Set)
  if type_ind = 0 then
    return vars
  end if
  for i=1 to length(File_Set) do
      if length(Toplevel_Variable_Sets[i]) >= type_ind then
        vars &= Toplevel_Variable_Sets[i][type_ind]
      end if
      if length(Type_Sets[i]) then
        for k=1 to length(Type_Sets[i]) do
	  if length(Type_Variable_Sets[i][k]) >= type_ind then
	    vars &= Type_Variable_Sets[i][k][type_ind]
	  end if
	end for
      end if
      if length(Function_Sets[i]) then
        for k=1 to length(Function_Sets[i]) do
	  if length(Function_Variable_Sets[i][k]) >= type_ind then
	    vars &= Function_Variable_Sets[i][k][type_ind]
	  end if
	end for
      end if
      if length(Procedure_Sets[i]) then
        for k=1 to length(Procedure_Sets[i]) do
	  if length(Procedure_Variable_Sets[i][k]) >= type_ind then
	    vars &= Procedure_Variable_Sets[i][k][type_ind]
	  end if
	end for
      end if
      if length(Global_Variable_Set) >= type_ind then
        vars &= Global_Variable_Set[type_ind]
      end if    
  end for
  return vars
end function  

--Someone else can finish these...
