--srcobj.e
--Creates a Source Object given a file name by paring that file.
--Also navigates a Source Object

include error.e

global constant SOURCE = 1,
		COMMENTS = 2
  
constant SYMBOL_CHARS = {'!', '~', '@', '$', '%', '^', '*', '-',
                         '=', '+', '/', 
			 '<', '>','|','\\', '.', '?'}
	 
--These symbols can't be grouped
constant ATOMIC_SYMBOLS = {'(',')','{','}','[',']',','}

integer line_num
sequence words, f_name, comments, c_comment

function get_string_const(integer wi,sequence line)
  integer c_char
  sequence word
  word = "\""
  wi += 1
  while wi <= length(line) do
    c_char=line[wi]
    if c_char = '\\' then --grab 2 chars
      word &= '\\'
      if length(line) > wi then
        wi += 1
	word &= line[wi]
      else
        parse_error("Incomplete string constant",
	            f_name,line_num,line,wi)
      end if
    elsif c_char = '\"' then
      word &= '\"'
      words &= {word}
      return wi + 1
    else
      word &= c_char
    end if
    wi += 1
  end while
  parse_error("Incomplete string constant", --end of line without end quote
              f_name,line_num,line,wi)
end function

function get_char_const(integer wi,sequence line)
  integer c_char
  sequence word
  word = "\'"
  wi += 1
  if wi < length(line) then
    c_char=line[wi]
    if c_char = '\\' then --grab 2 chars
      word &= '\\'
      if length(line) > wi + 1 then
        wi += 1
	word &= line[wi..wi+1]
	wi += 2     
      else
        parse_error("Incomplete char constant",
	            f_name,line_num,line,wi)
      end if
    else
      word &= c_char & "\'"
      wi += 2 
    end if
  else
    parse_error("Invalid char constant",
	        f_name,line_num,line,wi+1)
  end if
  if line[wi-1] != '\'' then
    parse_error("Invalid char constant",
	        f_name,line_num,line,wi+1)
  end if
  words &= {word}
  return wi
end function


function get_symbol(integer wi, sequence line)
  --modifies the global "words"
  integer c_char
  sequence word 
  word = {line[wi]}
  wi += 1
  while wi <= length(line) do
    c_char = line[wi]
    if find(c_char,SYMBOL_CHARS) then 
      word &= c_char
      wi += 1
    else
      words &= {word}
      return wi
    end if  
  end while 
  words &= {word}
  return wi
end function

function first_symbol_char(integer c)
  return find(c,SYMBOL_CHARS)
end function 

function get_atom(integer wi, sequence line)
  --modifies the global "words"
  integer c_char,radix_flag
  sequence word
  word = {line[wi]}
  radix_flag = line[wi] = '.'
  wi += 1
  while wi <= length(line) do
    c_char = line[wi]
    if c_char >= '0' and c_char <= '9' then
      word &= c_char
      wi += 1
    elsif c_char = '.' then
      if wi < length(line) then
        if find(line[wi+1],SYMBOL_CHARS) then --atom symbol
	  words &= {word}
	  return get_symbol(wi,line)
	end if
      end if 
      if radix_flag then --two decimal points
        parse_error("Unexpected char while parsing atom",
	            f_name, line_num, line, wi)
      else
        word &= c_char
	radix_flag = 1
        wi += 1
      end if	
    else
      exit
    end if  
  end while 
  --check error here too
  if word[length(word)] = '.' then
    parse_error("The atom " & word & " is invalid",
                f_name, line_num, line, wi)
  end if
  words &= {word}
  return wi
end function

function first_atom_char(integer c)
  return c >= '0' and c <= '9'
end function

 
function get_hex(integer wi, sequence line)
  --modifies the global "words"
  integer c_char
  sequence word
  word = {line[wi]}
  wi += 1
  while wi <= length(line) do
    c_char = line[wi]
    if (c_char <= '9' and c_char >= '0') or
       (c_char <= 'F' and c_char >= 'A') then
       word &= c_char
       wi += 1
    else
      exit
    end if
  end while    
  if length(word) > 1 then --OK
    words &= {word}
    return wi
  else
    parse_error("The hex " & word & " is invalid",
                f_name, line_num, line, wi)
  end if         
end function  

function first_hex_char(integer c)
  return c = '#'
end function

function get_word(integer wi, sequence line)
  --modifies the global "words"
  integer c_char
  sequence word
  word = {line[wi]}
  wi += 1
  while wi <= length(line) do
    c_char = line[wi]
    if (c_char <= '9' and c_char >= '0') or
       (c_char <= 'Z' and c_char >= 'A') or
       (c_char <= 'z' and c_char >= 'a') or
       c_char = '_' then
       word &= c_char
       wi += 1
    elsif c_char = '.' then
      if wi < length(line) then
	if find(line[wi+1],SYMBOL_CHARS) then --word symbol
	  words &= {word}
          return get_symbol(wi,line)
	else
	  word &= c_char
	  wi += 1
	end if
      end if 
    else
      exit
    end if
  end while    
  words &= {word}
  return wi
end function    

function atomic_symbol_char(integer c)
  return find(c,ATOMIC_SYMBOLS)
end function

function first_word_char(integer c)
  return (c >= 'a' and c <= 'z') or (c >= 'A' and c <= 'Z')
end function

global function New_Source_Object(sequence file_name)
  integer fileId,wi,cur_char
  sequence src
  object line
  --Try to open the file
  fileId = poe_open(file_name,"r")
  f_name = file_name --make available as global
  src = {}
  comments = {}
  --get the first line
  line_num = 1
  line = gets(fileId)
  while sequence(line) do --until EOF
    words = {}
    c_comment = ""
    line = line[1..length(line)-1] --strip CR
    if length(line) then
      wi = 1
      while wi <= length(line) do 
        cur_char = line[wi]
 	if first_word_char(cur_char) then
	  wi = get_word(wi,line)
	elsif first_atom_char(cur_char) then
	  wi = get_atom(wi,line)
      elsif atomic_symbol_char(cur_char) then
        words &= {{cur_char}}
        wi += 1  
	elsif first_symbol_char(cur_char) then
	  if wi < length(line) then
	    if cur_char = '.' and 
	       line[wi+1] >= '0' and line[wi+1] <= '9' then
	      wi = get_atom(wi,line)
            elsif cur_char = '-' and line[wi+1] = '-' then
	      c_comment = line[wi..length(line)]
	      exit
	    else
	      wi = get_symbol(wi,line)
	    end if
	  else
	    --*********************************
	    --Line end in , char
	    wi = get_symbol(wi,line)  -- a little patch
	    --I'LL HAVE TO LOOK AT THIS CASE AGAIN
	  end if
	elsif first_hex_char(cur_char) then
	  wi = get_hex(wi,line)
        elsif cur_char = '\"' then
	  wi = get_string_const(wi,line)
	elsif cur_char = '\'' then
	  wi = get_char_const(wi,line)  
	else
	  wi += 1  
	end if
      end while	    
    end if
    comments &= {c_comment}
    src &= {words}
    line = gets(fileId) --next line
    line_num += 1
  end while
  close(fileId)
  return {src,comments}
end function
